#pragma once
#include "Nodes/linkedlistnode.h"

namespace Datastructures {
	template <typename T> 
	class LinkedList
	{
	public:
		LinkedList();
		LinkedList(T value);
		~LinkedList();
		T operator[](int index);
		void append(T value);
		bool insert(T value, int index);
		bool remove(int index);
		int count() const;

	private:
		std::shared_ptr<Nodes::LinkedListNode<T>> getIndex(int index);
		std::shared_ptr<Nodes::LinkedListNode<T>> mHead;
		std::shared_ptr<Nodes::LinkedListNode<T>> mTail;
		int mCount;
	};

	//Generic Class constructor for the Linked List
	template <typename T>
	LinkedList<T>::LinkedList() {
		mCount = 0;
		mHead = nullptr;
		mTail = nullptr;
	}

	//Class constructor for the Linked List, taking a preset value to go into the head of the list
	template <typename T>
	LinkedList<T>::LinkedList(T value) {
		mCount = 1;
		mHead = std::make_shared<Nodes::LinkedListNode<T>>(value);
		mTail = mHead;
	}

	template <typename T>
	LinkedList<T>::~LinkedList() {}

	template <typename T>
	void LinkedList<T>::append(T value) {
		mCount++;
		if (mHead == nullptr) {
			mHead = std::make_shared<Nodes::LinkedListNode<T>>(value);
			mTail = mHead;
			return;
		}

		//Add new node and set to tail.
		mTail->next = std::make_shared<Nodes::LinkedListNode<T>>(value);
		mTail = mTail->next;
	}

	template <typename T>
	bool LinkedList<T>::insert(T value, int index) {

		if (index == 0 || this->mCount == 0)
		{//If the new item needs to go into the head of the linked list.
			mHead = std::make_shared<Nodes::LinkedListNode<T>>(value, mHead);
		} else {
			//Get the node before the one we are going to be inserting
			std::shared_ptr<Nodes::LinkedListNode<T>> node = this->getIndex(index-1);
			//If node is nullptr, index is out of range
			if (node == nullptr)
				return false;

			node->next = std::make_shared<Nodes::LinkedListNode<T>>(value, node->next);
		}

		//Append the new value into a new node.
		mCount++;
		return true;
	}

	template <typename T>
	std::shared_ptr<Nodes::LinkedListNode<T>> LinkedList<T>::getIndex(int index) {
		//Check if the value lies within the range of the list.
		if (index < 0 || index >= this->mCount)
			return nullptr;

		std::shared_ptr<Nodes::LinkedListNode<T>> node = mHead;
		for (size_t i = 0; i < index; i++)
		{//Interate through the linked list i times to get to the index.
			node = node->next;
		}
		return node;

	}
}